/*
                      cube.c
                      ======
                 Uwe Berger; 2007
                 
Aufgaben:
---------
-Hardwareansteuerung des 3D-Cube via Interrupt-Routine, welche Ebenen, Spalten
 und Zeilen hochzaehlt sowie Inhalte an Cube-Hardware ausgibt
-Kommunikationsprotokoll (via Ethernet), um hauptsaechlich die Bildinhalte des
 3D-Cube an Etherrrape zu uebertragen
-Remote-Control via IR-Fernbedienung


Hardwarekonfiguration:
----------------------
PA0 - PA2 -> Spaltenzaehler -> 8aus3-Decoder (74xx237x)
PA3 - PA5 -> Ebenenzaehler  -> 8aus3-Decoder (74xx237x)

PC4 -> SER \
PC5 -> SCK  > 8Bit-Schieberegister (74xx595x; SCL auf H; G auf L)
PC6 -> RCK /

Achtung: die hier eingestellten Timerwerte sind optimiert fuer einen 
         3x3x3-Wuerfel

*/



#include <string.h>
#include "cube.h"
#include "rc5.h"
#include "uart.h"
#include <avr/io.h>
#include <avr/interrupt.h>


/*  Timer */
volatile uint16_t timer_preload  = 65493; /* 1s->46005; 0,5s->55770; 225Hz->65449; 450Hz->65493 */
volatile uint8_t  timer_prescale = 0x05;  /* 1:1024                                      */

/*  ...   */
volatile uint8_t  ebene = 0;
volatile uint8_t  spalte = 0;
volatile uint8_t  led[CUBE_XYZ];

volatile uint8_t  rc_control = 0;
volatile uint8_t  rc_cursor  = 0;
volatile uint8_t  rc_led_pos = 0;


ISR(TIMER1_OVF_vect)
{
	
	TCNT1   =  timer_preload;        /* Preload-Wert     */
	
	/* Ebenen und Spalten durchzaehlen */
	if (spalte < CUBE_Y - 1) {
		spalte++;
	} else {
		if (ebene  < CUBE_Z - 1) {ebene++;}  else {ebene = 0;}
		spalte = 0;
	}
	PORTA = (ebene * 8 + spalte);

	/* eine Zeile ins Schieberegister schieben */
	for (uint8_t i = 0; i < CUBE_X; i++) {
		if (led[(ebene*CUBE_Z*CUBE_Y)+(spalte*CUBE_Y)+i]) {
			PORTC |= (1 << PC4);
		} else {
			PORTC &= ~(1 << PC4);
		}
		/* 0-1-Flanke an SCK schiebt Daten in den Registern um eine Stelle weiter */
		PORTC &= ~(1 << PC5); /* SCK auf 0 */
		PORTC |= (1 << PC5);  /* SCK auf 1 */
	}
	/* 0-1-Flanke an RCK bringt die Daten aus den Registern in die Latches */
	PORTC &= ~(1 << PC6); /* RCK auf 0 */
	PORTC |= (1 << PC6);  /* RCK auf 1 */
	
}


void SplitCmdArg(char * line, char ** cmd, char ** args)
{
    /* vorangestellte Leerzeichen ueberlesen */
    while (*line && *line <= ' ') {line++;}
    
    /* 1.Wort ist Kommando, in Grossbuchstaben konvertieren */
    *cmd = line;
    while (*line > ' ') {
        if (*line >= (unsigned char) 'a' && *line <= (unsigned char) 'z') { *line -= (unsigned char) 'a' - 'A';}
        line++;
    }
    
    /* Endemarkierung des Kommandowortes */
    if (*line) {*line++ = '\0';}
    
    /* Leerzeichen ueberlesen */
    while (*line && *line <= ' ') {++line;}
    
    /* dann kommen die Argumente */
    *args = line;
    while (*line && *line != '\r' && *line != '\n') {line++;}
    
    /* Endemarkierung der Argumente */
    *line = 0;
}



void cube_init(void)
{
	/* IP-Port */
	uip_listen(HTONS(CUBE_PORT));
	
	/* benoetigte Ports als Ausgaenge schalten */
	DDRA = (1 << DDA0) | (1 << DDA1) | (1 << DDA2) | (1 << DDA3) | (1 << DDA4) | (1 << DDA5);
	DDRC = (1 << DDC4) | (1 << DDC5) | (1 << DDC6);

	/* Timer und Interrupt initialisieren*/
	TCNT1   =  timer_preload;        /* Preload-Wert     */
	TCCR1B |=  timer_prescale;       /* Vorteiler      ) */
	TIMSK1 |= (1 << TOIE1);          /* Interrupt Timer1 */
	sei();
}


void set_all_led(uint8_t v)
{
	for (uint8_t i = 0; i < CUBE_XYZ; i++) {
		led[i] = v;
	}
}

void cube_rc5(void)
{
	uint8_t l = 0;
	uint8_t rc_cmd;
	uint8_t xx, zz;
	
	while (l < rc5_global.len) {
	
		if (rc_cursor == 1) {
			xx = (rc_led_pos/CUBE_X) + 1;
			zz = rc_led_pos/(CUBE_X * CUBE_Y);		
		}

		rc_cmd = rc5_global.queue[l].code;

		//uart_putdecbyte(rc_cmd);
		//uart_eol();
		
		switch (rc_cmd) {
		
		case RC_CMD_OFF:	rc_control = 0;
					rc_cursor  = 0;
					break;
		
		case RC_CMD_1:		rc_control = 1;
					rc_cursor  = 0;
					set_all_led(1);
					break;
		
		case RC_CMD_2:		rc_control = 1;
					rc_cursor  = 0;
					set_all_led(0);
					break;

		case RC_CMD_3:		rc_control = 1;
					rc_cursor  = 1;
					set_all_led(0);
					led[rc_led_pos] = 1;
					break;

		case RC_CMD_DOWN:	if (rc_cursor == 1) {
						if ((rc_led_pos + CUBE_Y) < ((zz+1) * CUBE_X * CUBE_Y)) {
							rc_led_pos = rc_led_pos + CUBE_Y;
						}
						set_all_led(0);
						led[rc_led_pos] = 1;
					}
					break;

		case RC_CMD_TOP:	if (rc_cursor == 1) {
						if ((rc_led_pos - CUBE_Y) >= (zz * CUBE_X * CUBE_Y)) {
							rc_led_pos = rc_led_pos - CUBE_Y;
						}
						set_all_led(0);
						led[rc_led_pos] = 1;
					}
					break;

		case RC_CMD_CH_PLUS:	if (rc_cursor == 1) {
						if ((rc_led_pos + CUBE_X*CUBE_Y) < CUBE_XYZ) {
							rc_led_pos = rc_led_pos + CUBE_X*CUBE_Y;
						}
						set_all_led(0);
						led[rc_led_pos] = 1;
					}
					break;

		case RC_CMD_CH_MINUS:	if (rc_cursor == 1) {
						if ((rc_led_pos - CUBE_X*CUBE_Y) >= 0) {
							rc_led_pos = rc_led_pos - CUBE_X*CUBE_Y;
						}
						set_all_led(0);
						led[rc_led_pos] = 1;
					}
					break;
					
		case RC_CMD_LEFT:	if (rc_cursor == 1) {
						if ((rc_led_pos - 1) >= ((xx-1)*CUBE_X)) {
							rc_led_pos = rc_led_pos - 1;
						}
						set_all_led(0);
						led[rc_led_pos] = 1;
					}
					break;

		case RC_CMD_RIGHT:	if (rc_cursor == 1) {
						if ((rc_led_pos + 1) < (xx*CUBE_X)) {
							rc_led_pos = rc_led_pos + 1;
						}
						set_all_led(0);
						led[rc_led_pos] = 1;
					}
					break;
		}
	
        l++;

	}

    /* clear queue */
    rc5_global.len = 0;

}

void cube_main(void)
{

	struct cube_cmd_state *s = &(uip_conn->appstate);

	char* cmd;
	char* arg;
	char s_set_led[10+CUBE_XYZ] = "set_led ";
	uint16_t arg_len;


	if(uip_rexmit())
		uip_send("cannot reximt. sorry.\n", 22);

	if(uip_newdata()) {
		
		/* empfangenes Kommando auseinandernehmen */
		SplitCmdArg(uip_appdata, &cmd, &arg);
		
		/* ein neues Kommando beginnt? */
		if (! strcmp(cmd, CMD_SET_LED)     ||
		    ! strcmp(cmd, CMD_GET_XYZ)     ||
		    ! strcmp(cmd, CMD_GET_LED)     ||
		    ! strcmp(cmd, CMD_VERSION)     ||
		    ! strcmp(cmd, CMD_RC5_CONTROL) ||
		    ! strcmp(cmd, CMD_SET_PRELOAD)   ) 
		{
			s->state = NEW_CMD;
			s->arg_pos = 0;
			s->textlen = 0;
		}
		
		/* Kommando send_led noch nicht vollstaendig? */
		if (s->state == CMD_SET_LED_RECEIVE) {
			/* cmd hat hier die Nutzdaten! */
			arg_len = strlen(cmd); 
			for (uint16_t i = 0; i < arg_len; i++) {
				if ((s->arg_pos + i) < CUBE_XYZ && rc_control == 0) {
					led[s->arg_pos + i] = cmd[i]-'0';
				}
			}
			s->arg_pos = s->arg_pos + arg_len; 
			/* fehlt immer noch was? */
			if (s->arg_pos < CUBE_XYZ) {
				s->state = CMD_SET_LED_RECEIVE;
			} else {
				s->textptr = out_update_ok;
				s->textlen = strlen(out_update_ok);
			}
		}

		/* Schaltzustaende der LEDs uebernehmen */
    		if (! strcmp(cmd, CMD_SET_LED)) {
			arg_len = strlen(arg); 
			for (uint16_t i = 0; i < arg_len; i++) {
				if (i < CUBE_XYZ && rc_control == 0) {
					led[i] = arg[i]-'0';
				}
			}
			s->arg_pos = arg_len;
			/* fehlt noch was? */
			if (s->arg_pos < CUBE_XYZ) {
				s->state = CMD_SET_LED_RECEIVE;
			} else {
				s->textptr = out_update_ok;
				s->textlen = strlen(out_update_ok);
			}
		}
		
    		/* Dimension des LED-Cube */
		if(! strcmp(cmd, CMD_GET_XYZ)) {
			s->textptr = out_set_xyz;
			s->textlen = strlen(out_set_xyz);
		}

    		/* Schaltzustaende der LEDs abfragen */
		if(! strcmp(cmd, CMD_GET_LED)) {
			for (uint16_t i = 0; i < CUBE_XYZ; i++) {
				if (led[i]) {
					strcat(s_set_led, "1");
				} else {
					strcat(s_set_led, "0");
				}
			}
			strcat(s_set_led, "\n");
			s->textptr = s_set_led;
			s->textlen = strlen(s_set_led);
		}
	
    		/* Versionsabfrage */
		if(! strcmp(cmd, CMD_VERSION)) {
			s->textptr = build_date;
			s->textlen = strlen(build_date);
		}

    		/* zum TEST ... Preload setzen */
		if(! strcmp(cmd, CMD_SET_PRELOAD)) {
			timer_preload = atoi(arg);
		}

    		/* zum TEST ... Abfrage Variable rc5_control */
		if(! strcmp(cmd, CMD_RC5_CONTROL)) {
			if (rc_control == 0) {
				s->textptr = "OFF\n";
				s->textlen = 4;
			} else {
				s->textptr = "ON\n";
				s->textlen = 3;
			}
		}

	}
	
	/* gibt es etwas zurueckzusenden? */
	if (uip_newdata()   ||
	    uip_rexmit()    ||
	    uip_acked()     ||
	    uip_poll()      ||
	    uip_connected())  {
		if (s->textlen > 0) {
			uip_send(s->textptr, s->textlen);
			s->textlen = 0;
		}
	}
	
}
