/* vim:fdm=marker ts=4 et ai
 * {{{
 *         simple rc5 implementation
 *
 *    for additional information please
 *    see http://lochraster.org/etherrape
 *
 * (c) by Alexander Neumann <alexander@bumpern.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * For more information on the GPL, please go to:
 * http://www.gnu.org/copyleft/gpl.html
 }}} */

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <string.h>

#include "config.h"
#include "rc5.h"

#define noinline __attribute__((noinline))



/* global variables */
volatile struct rc5_global_t rc5_global;

/* the rc5 bit length (= duration for one bit, encoded in rc5) is 1.8ms
 * (see atmel avr application note AVR410)
 * RC5_BIT_TICKS is the amount of timer ticks for one complete bit
 * RC5_HALF_BIT_TICKS is amount of timer ticks for half a bit
 *
 * (for cpu speed F_CPU and prescaler 1024) */

#if !F_CPU == 20000000UL
# error "please correct timing values in rc5/rc5.h!"
#endif

/* #define RC5_BIT_TICKS ((uint8_t)(F_CPU * 1800 / 1024 / 1000000)) */
#define RC5_BIT_TICKS 35    /* for 20MHz */

#define RC5_HALF_BIT_TICKS ((uint8_t)(RC5_BIT_TICKS/2))

/* define maximal environment for detecting valid rc5 pulses */
#define RC5_ENVIRONMENT ((uint8_t)(RC5_BIT_TICKS / 5))



/* local variables, temporary command buffer */
static struct rc5_t temp_rc5;

/* module local prototypes */
void noinline rc5_send_one(void);
void noinline rc5_send_zero(void);

void rc5_init(void)
{

    /* configure send pin as output, set low */
    RC5_SEND_DDR |= _BV(RC5_SEND_PINNUM);
    RC5_SEND_PORT &= ~_BV(RC5_SEND_PINNUM);

    /* enable timer0, set prescaler and enable overflow interrupt */
    TCCR0A = 0;
    TCCR0B = _BV(CS02) | _BV(CS00);

    /* configure int0 to fire at any logical change */
    EICRA |= _BV(ISC00);
    EICRA &= ~_BV(ISC01);

    /* clear any old interrupts and enable int0 interrupt */
    EIFR = _BV(INTF0);
    EIMSK |= _BV(INT0);

    /* reset everything to zero */
    memset((void *)&rc5_global, 0, sizeof(rc5_global));

    /* enable rc5 receive, init variables */
    rc5_global.enabled = 1;
}

void rc5_send(uint8_t addr, uint8_t cmd)
{
    static uint8_t toggle = 0;

    /* send two one sync bits */
    rc5_send_one();
    rc5_send_one();

    /* send toggle bit */
    if (toggle)
        rc5_send_one();
    else
        rc5_send_zero();

    toggle = !toggle;

    for (int8_t i = 4; i >= 0; i--) {
        if (addr & _BV(i))
            rc5_send_one();
        else
            rc5_send_zero();
    }

    for (int8_t i = 5; i >= 0; i--) {
        if (cmd & _BV(i))
            rc5_send_one();
        else
            rc5_send_zero();
    }

    /* turn off sender */
    RC5_SEND_PORT &= ~_BV(RC5_SEND_PINNUM);

}

void rc5_send_one(void)
{
    RC5_SEND_PORT &= ~_BV(RC5_SEND_PINNUM);
    _delay_loop_2(RC5_PULSE);
    RC5_SEND_PORT |= _BV(RC5_SEND_PINNUM);
    _delay_loop_2(RC5_PULSE);
}

void rc5_send_zero(void)
{
    RC5_SEND_PORT |= _BV(RC5_SEND_PINNUM);
    _delay_loop_2(RC5_PULSE);
    RC5_SEND_PORT &= ~_BV(RC5_SEND_PINNUM);
    _delay_loop_2(RC5_PULSE);
}

void rc5_process(void)
{
    static uint8_t toggle = 2;

    if (rc5_global.new_data) {

        if (toggle != rc5_global.received_command.toggle_bit) {

            /* shift queue backwards */
            memmove(&rc5_global.queue[1],
                    &rc5_global.queue[0],
                    (RC5_QUEUE_LENGTH-1) * sizeof(struct rc5_t));

            /* copy datagram to queue and increment length */
            rc5_global.queue[0].raw = rc5_global.received_command.raw;
            if (rc5_global.len < RC5_QUEUE_LENGTH)
                rc5_global.len++;

            toggle = rc5_global.received_command.toggle_bit;

        }
        rc5_global.new_data = 0;
    }
}

/* int0 interrupt */
ISR(INT0_vect)
{

    if (rc5_global.enabled && !rc5_global.temp_disable) {

        /* if this is the first interrupt */
        if (rc5_global.interrupts == 0) {
            /* reset counter, clear old overflows and enable
             * timer0 overflow interrupt */
            TCNT0 = 0;
            TIFR0 = _BV(TOV0);
            TIMSK0 |= _BV(TOIE0);

            /* reset temp buffer */
            temp_rc5.raw = 0;

        /* if this is not the first interrupt */
        } else {
            /* load and reset the counter */
            uint8_t counter = TCNT0;
            TCNT0 = 0;

            /* check how many halfbits have passed since last interrupt */
            uint8_t received_bits = 0;

            /* check for two halfbits */
            if ( (counter > RC5_HALF_BIT_TICKS - RC5_ENVIRONMENT &&
                  counter < RC5_HALF_BIT_TICKS + RC5_ENVIRONMENT) ) {
                received_bits = 1;

            /* check for one halfbit */
            } else if ( (counter > RC5_BIT_TICKS - RC5_ENVIRONMENT &&
                         counter < RC5_BIT_TICKS + RC5_ENVIRONMENT) ) {
                received_bits = 2;

            /* else signal is invalid */
            } else {
                /* disable interrupt, enable timer overflow interrupt: -> timeout */
                //GICR &= ~_BV(INT0);
                rc5_global.temp_disable = 1;

                /* reset bitcounter */
                rc5_global.halfbitcount = 0;

                /* quit this interrupt and wait for timeout */
                return;
            }

            /* process each received halfbit */
            while (received_bits--) {
                /* increase halfbit counter, since we obviously
                 * received a halfbit */
                rc5_global.halfbitcount++;

                /* if the parity of the halfbit-counter is odd,
                 * we have received a complete bit,
                 * so save this bit to the structure*/

                if (rc5_global.halfbitcount & 1) {
                    /* shift buffer left, and set lsb, if the
                     * parity of the interrupt counter (interrupts
                     * which happened _BEFORE_ this one (=postincrement))
                     * is odd*/
                    temp_rc5.raw <<= 1;
                    temp_rc5.raw |= (rc5_global.interrupts & 1);
                }
            }
        }

        /* increase interrupt counter */
        rc5_global.interrupts++;
    }
}

/* timer0 overflow interrupt */
ISR(TIMER0_OVF_vect)
{

    /* disable overflow interrupt */
    TIMSK0 &= ~_BV(TOIE0);

    /* check if we only received 26 halfbits,
     * so the last transmitted bit was zero,
     * if 27 halfbits have been received, everything
     * went fine */
    switch (rc5_global.halfbitcount) {
        case 26:
                 /* add a zero */
                 temp_rc5.raw <<= 1;
                 /* and fall-through to the next case */

        case 27:
                 /* copy data to global structure */
                 rc5_global.received_command.raw = temp_rc5.raw;

                 /* signal main that new data has arrived */
                 rc5_global.new_data = 1;
    }

    /* reset counter */
    rc5_global.halfbitcount = 0;
    rc5_global.interrupts = 0;

    /* if decoder is enabled, reconfigure int0 */
    if (rc5_global.enabled) {

        /* re-enable int0 */
        rc5_global.temp_disable = 0;

    }

}


